@@ -28,6 +28,7 @@ import re
 import shlex
 import shutil
 import socket
+import socketserver
 import stat
 import struct
 import subprocess
@@ -35,18 +36,14 @@ import sys
 import tempfile
 import threading
 import time
+from http.server import HTTPServer, SimpleHTTPRequestHandler
 from operator import itemgetter
+from urllib.parse import unquote, urlsplit
 
-if sys.version_info.major == 2:
-  import SocketServer as socketserver
-  from BaseHTTPServer import HTTPServer
-  from SimpleHTTPServer import SimpleHTTPRequestHandler
-  from urllib import unquote
-  from urlparse import urlsplit
-else:
-  import socketserver
-  from http.server import HTTPServer, SimpleHTTPRequestHandler
-  from urllib.parse import unquote, urlsplit
+# We depend on python 3.8 features
+if sys.version_info < (3, 8): # noqa: UP036
+  print(f'error: emrun requires python 3.8 or above ({sys.executable} {sys.version})', file=sys.stderr)
+  sys.exit(1)
 
 # Populated from cmdline params
 emrun_options = None
@@ -86,7 +83,7 @@ page_exit_code = None
 # launcher.exe that runs the actual opera browser.  So killing browser_process
 # would just kill launcher.exe and not the opera
 # browser itself.
-processname_killed_atexit = ""
+processname_killed_atexit = ''
 
 # Using "0.0.0.0" means "all interfaces", which should allow connecting to this
 # server via LAN addresses. Using "localhost" should allow only connecting from
@@ -120,8 +117,6 @@ if not WINDOWS and not LINUX and not MACOS:
 
 # Returns wallclock time in seconds.
 def tick():
-  # Would like to return time.clock() since it's apparently better for
-  # precision, but it is broken on macOS 10.10 and Python 2.7.8.
   return time.time()
 
 
@@ -369,10 +364,10 @@ def kill_browser_process():
   global browser_process, processname_killed_atexit, current_browser_processes
   if browser_process and browser_process.poll() is None:
     try:
-      logv('Terminating browser process pid=' + str(browser_process.pid) + '..')
+      logv(f'Terminating browser process pid={browser_process.pid}..')
       browser_process.kill()
     except Exception as e:
-      logv('Failed with error ' + str(e) + '!')
+      logv(f'Failed with error {e}!')
 
     browser_process = None
     # We have a hold of the target browser process explicitly, no need to resort to killall,
@@ -382,17 +377,17 @@ def kill_browser_process():
   if current_browser_processes:
     for pid in current_browser_processes:
       try:
-        logv('Terminating browser process pid=' + str(pid['pid']) + '..')
+        logv(f'Terminating browser process pid={pid["pid"]}..')
         os.kill(pid['pid'], 9)
       except Exception as e:
-        logv('Failed with error ' + str(e) + '!')
+        logv(f'Failed with error {e}!')
 
     current_browser_processes = None
     # We have a hold of the target browser process explicitly, no need to resort to killall,
     # so clear that record out.
     processname_killed_atexit = ''
 
-  if len(processname_killed_atexit):
+  if processname_killed_atexit:
     if emrun_options.android:
       logv("Terminating Android app '" + processname_killed_atexit + "'.")
       subprocess.call([ADB, 'shell', 'am', 'force-stop', processname_killed_atexit])
@@ -439,7 +434,7 @@ def detect_browser_processes():
     return False
 
   for p in running_browser_processes:
-    logv('Detected running browser process id: ' + str(p['pid']) + ', existed already at emrun startup? ' + str(pid_existed(p['pid'])))
+    logv(f'Detected running browser process id: {p["pid"]}, existed already at emrun startup? {pid_existed(p["pid"])}')
 
   current_browser_processes = [p for p in running_browser_processes if not pid_existed(p['pid'])]
 
@@ -544,7 +539,7 @@ class HTTPWebServer(socketserver.ThreadingMixIn, HTTPServer):
       time_since_message = now - last_message_time
       if emrun_options.silence_timeout != 0 and time_since_message > emrun_options.silence_timeout:
         self.shutdown()
-        logi('No activity in ' + str(emrun_options.silence_timeout) + ' seconds. Quitting web server with return code ' + str(emrun_options.timeout_returncode) + '. (--silence-timeout option)')
+        logi(f'No activity in {emrun_options.silence_timeout} seconds. Quitting web server with return code {emrun_options.timeout_returncode}. (--silence-timeout option)')
         page_exit_code = emrun_options.timeout_returncode
         emrun_options.kill_exit = True
 
@@ -552,7 +547,7 @@ class HTTPWebServer(socketserver.ThreadingMixIn, HTTPServer):
       time_since_start = now - page_start_time
       if emrun_options.timeout != 0 and time_since_start > emrun_options.timeout:
         self.shutdown()
-        logi('Page has not finished in ' + str(emrun_options.timeout) + ' seconds. Quitting web server with return code ' + str(emrun_options.timeout_returncode) + '. (--timeout option)')
+        logi(f'Page has not finished in {emrun_options.timeout} seconds. Quitting web server with return code {emrun_options.timeout_returncode}. (--timeout option)')
         emrun_options.kill_exit = True
         page_exit_code = emrun_options.timeout_returncode
 
@@ -583,8 +578,6 @@ class HTTPWebServer(socketserver.ThreadingMixIn, HTTPServer):
 
 # Processes HTTP request back to the browser.
 class HTTPHandler(SimpleHTTPRequestHandler):
-  protocol_version = 'HTTP/1.1'  # noqa: DC01
-
   def send_head(self):
     global page_last_served_time
     path = self.translate_path(self.path)
@@ -679,7 +672,7 @@ class HTTPHandler(SimpleHTTPRequestHandler):
     global page_exit_code, have_received_messages
 
     (_, _, path, query, _) = urlsplit(self.path)
-    logv('POST: "' + self.path + '" (path: "' + path + '", query: "' + query + '")')
+    logv(f'POST: "{self.path}" (path: "{path}", query: "{query}")')
     if query.startswith('file='):
       # Binary file dump/upload handling. Requests to
       # "stdio.html?file=filename" will write binary data to the given file.
@@ -692,7 +685,7 @@ class HTTPHandler(SimpleHTTPRequestHandler):
         pass
       with open(filename, 'wb') as fh:
         fh.write(data)
-      logi('Wrote ' + str(len(data)) + ' bytes to file "' + filename + '".')
+      logi(f'Wrote {len(data)} bytes to file "{filename}".')
       have_received_messages = True
     elif path == '/system_info':
       system_info = json.loads(get_system_info(format_json=True))
@@ -721,7 +714,7 @@ class HTTPHandler(SimpleHTTPRequestHandler):
       elif data.startswith('^exit^'):
         if not emrun_options.serve_after_exit:
           page_exit_code = int(data[6:])
-          logv('Web page has quit with a call to exit() with return code ' + str(page_exit_code) + '. Shutting down web server. Pass --serve-after-exit to keep serving even after the page terminates with exit().')
+          logv(f'Web page has quit with a call to exit() with return code ${page_exit_code}. Shutting down web server. Pass --serve-after-exit to keep serving even after the page terminates with exit().')
           # Set server socket to nonblocking on shutdown to avoid sporadic deadlocks
           self.server.socket.setblocking(False)
           self.server.shutdown()
@@ -757,12 +750,8 @@ class HTTPHandler(SimpleHTTPRequestHandler):
 
 
 # Returns stdout by running command with universal_newlines=True
-def check_output(cmd, universal_newlines=True, *args, **kwargs):
-  if hasattr(subprocess, "run"):
-    return subprocess.run(cmd, universal_newlines=universal_newlines, stdout=subprocess.PIPE, check=True, *args, **kwargs).stdout
-  else:
-    # check_output is considered as an old API so prefer subprocess.run if possible
-    return subprocess.check_output(cmd, universal_newlines=universal_newlines, *args, **kwargs)
+def check_output(cmd, *args, **kwargs):
+  return subprocess.run(cmd, text=True, stdout=subprocess.PIPE, check=True, *args, **kwargs).stdout
 
 
 # From http://stackoverflow.com/questions/4842448/getting-processor-information-in-python
@@ -798,7 +787,7 @@ def get_cpu_info():
   except Exception as e:
     import traceback
     loge(traceback.format_exc())
-    return {'model': 'Unknown ("' + str(e) + '")',
+    return {'model': f'Unknown ("{e}")',
             'physicalCores': 1,
             'logicalCores': 1,
             'frequency': 0,
@@ -822,7 +811,7 @@ def get_android_cpu_infoline():
       hardware = line[line.find(':') + 1:].strip()
 
   freq = int(check_output([ADB, 'shell', 'cat', '/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq']).strip()) // 1000
-  return 'CPU: ' + processor + ', ' + hardware + ' @ ' + str(freq) + ' MHz'
+  return f'CPU: {processor}, {hardware} @ {freq} MHz'
 
 
 def win_get_gpu_info():
@@ -861,7 +850,7 @@ def win_get_gpu_info():
 
       try:
         driverDate = winreg.QueryValueEx(hVideoCardReg, 'DriverDate')[0]
-        VideoCardDescription += ' (' + driverDate + ')'
+        VideoCardDescription += f' ({driverDate})'
       except WindowsError:
         pass
 
@@ -1446,7 +1435,7 @@ def list_processes_by_name(exe_full_path):
     # Fail gracefully if psutil not available
     logv('import psutil failed, unable to detect browser processes')
 
-  logv('Searching for processes by full path name "' + exe_full_path + '".. found ' + str(len(pids)) + ' entries')
+  logv(f'Searching for processes by full path name "{exe_full_path}".. found {len(pids)} entries')
 
   return pids
 
@@ -1692,7 +1681,7 @@ def run(args):  # noqa: C901, PLR0912, PLR0915
     else:
       hostname = options.hostname
     # create url for browser after opening the server so we have the final port number in case we are binding to port 0
-    url = 'http://' + hostname + ':' + str(options.port) + '/' + url
+    url = f'http://{hostname}:{options.port}/{url}'
 
   if options.android:
     if options.run_browser or options.browser_info:
@@ -1726,7 +1715,7 @@ def run(args):  # noqa: C901, PLR0912, PLR0915
       # 5. Locate the name of the main activity for the browser in manifest.txt and add an entry to above list in form 'appname/mainactivityname'
 
       if options.android_tunnel:
-        subprocess.check_call([ADB, 'reverse', 'tcp:' + str(options.port), 'tcp:' + str(options.port)])
+        subprocess.check_call([ADB, 'reverse', f'tcp:{options.port}', f'tcp:{options.port}'])
 
       url = url.replace('&', '\\&')
       browser = [ADB, 'shell', 'am', 'start', '-a', 'android.intent.action.VIEW', '-n', browser_app, '-d', url]
@@ -1738,7 +1727,7 @@ def run(args):  # noqa: C901, PLR0912, PLR0915
     if options.run_browser or options.browser_info:
       browser = find_browser(str(options.browser))
       if not browser:
-        loge('Unable to find browser "' + str(options.browser) + '"! Check the correctness of the passed --browser=xxx parameter!')
+        loge(f'Unable to find browser "{options.browser}"! Check the correctness of the passed --browser=xxx parameter!')
         return 1
       browser_exe = browser[0]
       browser_args = shlex.split(unwrap(options.browser_args))
@@ -1794,7 +1783,7 @@ def run(args):  # noqa: C901, PLR0912, PLR0915
       run(['adb', 'shell', 'mkdir', '/mnt/sdcard/safe_firefox_profile'])
       run(['adb', 'push', os.path.join(profile_dir, 'prefs.js'), '/mnt/sdcard/safe_firefox_profile/prefs.js'])
     except Exception as e:
-      loge('Creating Firefox profile prefs.js file to internal storage in /mnt/sdcard failed with error ' + str(e) + '!')
+      loge(f'Creating Firefox profile prefs.js file to internal storage in /mnt/sdcard failed with error {e}!')
       loge('Try running without --safe-firefox-profile flag if unattended execution mode is not important, or')
       loge('enable rooted debugging on the Android device to allow adb to write files to /mnt/sdcard.')
     browser += ['--es', 'args', '"--profile /mnt/sdcard/safe_firefox_profile"']
@@ -1844,9 +1833,9 @@ def run(args):  # noqa: C901, PLR0912, PLR0915
       logv(browser_exe)
       previous_browser_processes = list_processes_by_name(browser_exe)
       for p in previous_browser_processes:
-        logv('Before spawning web browser, found a running ' + os.path.basename(browser_exe) + ' browser process id: ' + str(p['pid']))
+        logv(f'Before spawning web browser, found a running {os.path.basename(browser_exe)} browser process id: {p["pid"]}')
     browser_process = subprocess.Popen(browser, env=subprocess_env())
-    logv('Launched browser process with pid=' + str(browser_process.pid))
+    logv(f'Launched browser process with pid={browser_process.pid}')
     if options.kill_exit:
       atexit.register(kill_browser_process)
     # For Android automation, we execute adb, so this process does not
@@ -1858,7 +1847,7 @@ def run(args):  # noqa: C901, PLR0912, PLR0915
     premature_quit_code = browser_process.poll()
     if premature_quit_code is not None:
       options.serve_after_close = True
-      logv('Warning: emrun got immediately detached from the target browser process (the process quit with exit code ' + str(premature_quit_code) + '). Cannot detect when user closes the browser. Behaving as if --serve-after-close was passed in.')
+      logv(f'Warning: emrun got immediately detached from the target browser process (the process quit with exit code {premature_quit_code}). Cannot detect when user closes the browser. Behaving as if --serve-after-close was passed in.')
       if not options.browser:
         logv('Try passing the --browser=/path/to/browser option to avoid this from occurring. See https://github.com/emscripten-core/emscripten/issues/3234 for more discussion.')
 
@@ -1895,9 +1884,9 @@ def run(args):  # noqa: C901, PLR0912, PLR0915
 
 def main(args):
   returncode = run(args)
-  logv('emrun quitting with process exit code ' + str(returncode))
+  logv(f'emrun quitting with process exit code {returncode}')
   if temp_firefox_profile_dir is not None:
-    logi('Warning: Had to leave behind a temporary Firefox profile directory ' + temp_firefox_profile_dir + ' because --safe-firefox-profile was set and the browser did not quit before emrun did.')
+    logi(f'Warning: Had to leave behind a temporary Firefox profile directory {temp_firefox_profile_dir} because --safe-firefox-profile was set and the browser did not quit before emrun did.')
   return returncode
 
 
