@@ -19,14 +19,16 @@ http://kripken.github.io/emscripten-site/docs/getting_started/test-suite.html
 
 import argparse
 import atexit
+import datetime
 import fnmatch
 import glob
 import logging
 import math
 import operator
 import os
-import platform
 import random
+import re
+import subprocess
 import sys
 import time
 import unittest
@@ -37,14 +39,15 @@ from functools import cmp_to_key
 __rootpath__ = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
 sys.path.insert(0, __rootpath__)
 
+import browser_common
 import common
 import jsrun
 import parallel_testsuite
+from color_runner import ColorTextRunner
 from common import errlog
+from single_line_runner import SingleLineTestRunner
 
-from tools import config, shared, utils
-
-sys.path.append(utils.path_from_root('third_party/websockify'))
+from tools import building, colored_logger, config, shared, utils
 
 logger = logging.getLogger("runner")
 
@@ -59,6 +62,12 @@ passing_core_test_modes = [
   'core3',
   'cores',
   'corez',
+  'lto0',
+  'lto1',
+  'lto2',
+  'lto3',
+  'ltos',
+  'ltoz',
   'core_2gb',
   'strict',
   'strict_js',
@@ -415,9 +424,7 @@ def run_tests(options, suites):
     print('Test suites:', [s[0] for s in suites])
   # Run the discovered tests
 
-  # We currently don't support xmlrunner on macOS M1 runner since
-  # `pip` doesn't seeem to yet have pre-built binaries for M1.
-  if os.getenv('CI') and not (utils.MACOS and platform.machine() == 'arm64'):
+  if os.getenv('CI'):
     os.makedirs('out', exist_ok=True)
     # output fd must remain open until after testRunner.run() below
     output = open('out/test-results.xml', 'wb')
@@ -425,8 +432,16 @@ def run_tests(options, suites):
     testRunner = xmlrunner.XMLTestRunner(output=output, verbosity=2,
                                          failfast=options.failfast)
     print('Writing XML test output to ' + os.path.abspath(output.name))
+  elif options.ansi and not options.verbose:
+    # When not in verbose mode and ansi color output is available use our nice single-line
+    # result display.
+    testRunner = SingleLineTestRunner(failfast=options.failfast)
   else:
-    testRunner = unittest.TextTestRunner(verbosity=2, failfast=options.failfast)
+    if not options.ansi:
+      print('using verbose test runner (ANSI not avilable)')
+    else:
+      print('using verbose test runner (verbose output requested)')
+    testRunner = ColorTextRunner(failfast=options.failfast)
 
   total_core_time = 0
   run_start_time = time.perf_counter()
@@ -464,7 +479,12 @@ def parse_args():
                            'test.  Implies --cores=1.  Defaults to true when running a single test')
   parser.add_argument('--no-clean', action='store_true',
                       help='Do not clean the temporary directory before each test run')
-  parser.add_argument('--verbose', '-v', action='store_true')
+  parser.add_argument('--verbose', '-v', action='count', default=0,
+                      help="Show test stdout and stderr, and don't use the single-line test reporting. "
+                           'Specifying `-v` twice will enable test framework logging (i.e. EMTEST_VERBOSE)')
+  # TODO: Replace with BooleanOptionalAction once we can depend on python3.9
+  parser.add_argument('--ansi', action='store_true', default=None)
+  parser.add_argument('--no-ansi', action='store_false', dest='ansi', default=None)
   parser.add_argument('--all-engines', action='store_true')
   parser.add_argument('--detect-leaks', action='store_true')
   parser.add_argument('--skip-slow', action='store_true', help='Skip tests marked as slow')
@@ -489,6 +509,7 @@ def parse_args():
                            'Useful when combined with --failfast')
   parser.add_argument('--force64', action='store_true')
   parser.add_argument('--crossplatform-only', action='store_true')
+  parser.add_argument('--log-test-environment', action='store_true', help='Prints out detailed information about the current environment. Useful for adding more info to CI test runs.')
   parser.add_argument('--force-browser-process-termination', action='store_true', help='If true, a fail-safe method is used to ensure that all browser processes are terminated before and after the test suite run. Note that this option will terminate all browser processes, not just those launched by the harness, so will result in loss of all open browsing sessions.')
   parser.add_argument('--repeat', type=int, default=1,
                       help='Repeat each test N times (default: 1).')
@@ -496,6 +517,14 @@ def parse_args():
 
   options = parser.parse_args()
 
+  if options.ansi is None:
+    options.ansi = colored_logger.ansi_color_available()
+  else:
+    if options.ansi:
+      colored_logger.enable(force=True)
+    else:
+      colored_logger.disable()
+
   if options.failfast:
     if options.max_failures != 2**31 - 1:
       utils.exit_with_error('--failfast and --max-failures are mutually exclusive!')
@@ -505,9 +534,9 @@ def parse_args():
 
 
 def configure():
-  common.EMTEST_BROWSER = os.getenv('EMTEST_BROWSER')
-  common.EMTEST_BROWSER_AUTO_CONFIG = os.getenv('EMTEST_BROWSER_AUTO_CONFIG')
-  common.EMTEST_HEADLESS = int(os.getenv('EMTEST_HEADLESS', '0'))
+  browser_common.EMTEST_BROWSER = os.getenv('EMTEST_BROWSER')
+  browser_common.EMTEST_BROWSER_AUTO_CONFIG = os.getenv('EMTEST_BROWSER_AUTO_CONFIG')
+  browser_common.EMTEST_HEADLESS = int(os.getenv('EMTEST_HEADLESS', '0'))
   common.EMTEST_DETECT_TEMPFILE_LEAKS = int(os.getenv('EMTEST_DETECT_TEMPFILE_LEAKS', '0'))
   common.EMTEST_ALL_ENGINES = int(os.getenv('EMTEST_ALL_ENGINES', '0'))
   common.EMTEST_SKIP_SLOW = int(os.getenv('EMTEST_SKIP_SLOW', '0'))
@@ -522,7 +551,7 @@ def configure():
   assert 'PARALLEL_SUITE_EMCC_CORES' not in os.environ, 'use EMTEST_CORES rather than PARALLEL_SUITE_EMCC_CORES'
   parallel_testsuite.NUM_CORES = os.environ.get('EMTEST_CORES') or os.environ.get('EMCC_CORES')
 
-  common.configure_test_browser()
+  browser_common.configure_test_browser()
 
 
 def cleanup_emscripten_temp():
@@ -538,19 +567,98 @@ def cleanup_emscripten_temp():
         pass
 
 
+def print_repository_info(directory, repository_name):
+  current_commit = utils.run_process(['git', 'log', '--abbrev-commit', '-n1', '--pretty=oneline'], cwd=directory, stdout=subprocess.PIPE).stdout.strip()
+  print(f'\n{repository_name} {current_commit}\n')
+  local_changes = utils.run_process(['git', 'diff'], cwd=directory, stdout=subprocess.PIPE).stdout.strip()
+  if local_changes:
+    print(f'\n{local_changes}\n')
+
+
+def log_test_environment():
+  """Print detailed information about the current test environment. Useful for
+  logging test run configuration in a CI."""
+  print('======================== Test Setup ========================')
+  print(f'Test time: {datetime.datetime.now(datetime.timezone.utc).strftime("%A, %B %d, %Y %H:%M:%S %Z")}')
+  print(f'Python: "{sys.executable}". Version: {sys.version}')
+  print(f'Emscripten test runner path: "{os.path.realpath(__file__)}"')
+
+  if os.path.isdir(utils.path_from_root('.git')):
+    print(f'\nEmscripten repository: "{__rootpath__}"')
+
+  emscripten_version = utils.path_from_root('emscripten-version.txt')
+  if os.path.isfile(emscripten_version):
+    print(f'emscripten-version.txt: {utils.EMSCRIPTEN_VERSION}')
+
+  if os.path.isdir(os.path.join(__rootpath__, '.git')):
+    print_repository_info(__rootpath__, 'Emscripten')
+
+  print(f'EM_CONFIG: "{config.EM_CONFIG}"')
+  if os.path.isfile(config.EM_CONFIG):
+    print(f'\n{utils.read_file(config.EM_CONFIG).strip()}\n')
+
+  node_js_version = utils.run_process(config.NODE_JS + ['--version'], stdout=subprocess.PIPE).stdout.strip()
+  print(f'NODE_JS: {config.NODE_JS}. Version: {node_js_version}')
+
+  print(f'BINARYEN_ROOT: {config.BINARYEN_ROOT}')
+  wasm_opt_version = building.get_binaryen_version(building.get_binaryen_bin()).strip()
+  print(f'wasm-opt version: {wasm_opt_version}')
+
+  binaryen_git_dir = config.BINARYEN_ROOT
+  # Detect emsdk directory structure (build root vs source root)
+  if re.match(r'main_.*_64bit_binaryen', os.path.basename(binaryen_git_dir)):
+    binaryen_git_dir = os.path.realpath(os.path.join(binaryen_git_dir, '..', 'main'))
+  if os.path.isdir(os.path.join(binaryen_git_dir, '.git')):
+    print(f'Binaryen git directory: "{binaryen_git_dir}"')
+    print_repository_info(binaryen_git_dir, 'Binaryen')
+
+  print(f'LLVM_ROOT: {config.LLVM_ROOT}')
+
+  # Find LLVM git directory in emsdk aware fashion
+  def find_llvm_git_root(dir):
+    while True:
+      if os.path.isdir(os.path.join(dir, ".git")):
+        return dir
+      if os.path.isdir(os.path.join(dir, "src", ".git")):
+        return os.path.join(dir, "src")
+      if os.path.dirname(dir) == dir:
+        return None
+      dir = os.path.dirname(dir)
+
+  llvm_git_root = find_llvm_git_root(config.LLVM_ROOT)
+  if llvm_git_root:
+    print(f'LLVM git directory: "{llvm_git_root}"')
+    print_repository_info(llvm_git_root, 'LLVM')
+
+  clang_version = utils.run_process([shared.CLANG_CC, '--version'], stdout=subprocess.PIPE).stdout.strip()
+  print(f'Clang: "{shared.CLANG_CC}"\n{clang_version}\n')
+
+  print(f'EMTEST_BROWSER: {browser_common.EMTEST_BROWSER}')
+  if browser_common.is_firefox():
+    print(f'Firefox version: {browser_common.get_firefox_version()}')
+  else:
+    print('Not detected as a Firefox browser')
+  if browser_common.is_safari():
+    print(f'Safari version: {browser_common.get_safari_version()}')
+  else:
+    print('Not detected as a Safari browser')
+  if browser_common.is_chrome():
+    print('Browser is Chrome.')
+  else:
+    print('Not detected as a Chrome browser')
+
+  emsdk_dir = os.getenv('EMSDK')
+  print(f'\nEMSDK: "{emsdk_dir}"')
+  if emsdk_dir:
+    if os.path.isdir(os.path.join(emsdk_dir, '.git')):
+      print_repository_info(emsdk_dir, 'Emsdk')
+
+  print('==================== End of Test Setup =====================')
+
+
 def main():
   options = parse_args()
 
-  # Some options make sense being set in the environment, others not-so-much.
-  # TODO(sbc): eventually just make these command-line only.
-  if os.getenv('EMTEST_SAVE_DIR'):
-    errlog('ERROR: use --save-dir instead of EMTEST_SAVE_DIR=1, and --no-clean instead of EMTEST_SAVE_DIR=2')
-    return 1
-  if os.getenv('EMTEST_REBASELINE'):
-    errlog('Prefer --rebaseline over setting $EMTEST_REBASELINE')
-  if os.getenv('EMTEST_VERBOSE'):
-    errlog('Prefer --verbose over setting $EMTEST_VERBOSE')
-
   # We set the environments variables here and then call configure,
   # to apply them.  This means the python's multiprocessing child
   # process will see the same configuration even though they don't
@@ -577,7 +685,7 @@ def main():
   set_env('EMTEST_SKIP_SLOW', options.skip_slow)
   set_env('EMTEST_ALL_ENGINES', options.all_engines)
   set_env('EMTEST_REBASELINE', options.rebaseline)
-  set_env('EMTEST_VERBOSE', options.verbose)
+  set_env('EMTEST_VERBOSE', options.verbose > 1)
   set_env('EMTEST_CORES', options.cores)
   set_env('EMTEST_FORCE64', options.force64)
 
@@ -588,20 +696,11 @@ def main():
   # Remove any old test files before starting the run
   cleanup_emscripten_temp()
   utils.delete_file(common.flaky_tests_log_filename)
-  utils.delete_file(common.browser_spawn_lock_filename)
-  utils.delete_file(f'{common.browser_spawn_lock_filename}_counter')
-  if options.force_browser_process_termination or os.getenv('EMTEST_FORCE_BROWSER_PROCESS_TERMINATION'):
-    config = common.get_browser_config()
 
-    def terminate_all_browser_processes():
-      procs = common.list_processes_by_name(config.executable_name)
-      if len(procs) > 0:
-        print(f'Terminating {len(procs)} stray browser processes.')
-        common.terminate_list_of_processes(procs)
+  browser_common.init(options.force_browser_process_termination)
 
-    if config and hasattr(config, 'executable_name'):
-      atexit.register(terminate_all_browser_processes)
-      terminate_all_browser_processes()
+  if options.log_test_environment or os.getenv('CI'):
+    log_test_environment()
 
   def prepend_default(arg):
     if arg.startswith('test_'):
