@@ -112,18 +112,17 @@ def update_settings_glue(wasm_file, metadata, base_metadata):
     settings.WASM_EXPORTS = base_metadata.all_exports
   else:
     settings.WASM_EXPORTS = metadata.all_exports
-  settings.DATA_EXPORTS = list(metadata.data_exports.keys())
   settings.HAVE_EM_ASM = bool(settings.MAIN_MODULE or len(metadata.em_asm_consts) != 0)
 
   # start with the MVP features, and add any detected features.
-  settings.BINARYEN_FEATURES = ['--mvp-features'] + metadata.features
+  building.binaryen_features = ['--mvp-features'] + metadata.features
   if settings.ASYNCIFY == 2:
-    settings.BINARYEN_FEATURES += ['--enable-reference-types']
+    building.binaryen_features += ['--enable-reference-types']
 
   if settings.PTHREADS:
-    assert '--enable-threads' in settings.BINARYEN_FEATURES
+    assert '--enable-threads' in building.binaryen_features
   if settings.MEMORY64:
-    assert '--enable-memory64' in settings.BINARYEN_FEATURES
+    assert '--enable-memory64' in building.binaryen_features
 
   settings.HAS_MAIN = bool(settings.MAIN_MODULE) or settings.PROXY_TO_PTHREAD or settings.STANDALONE_WASM or 'main' in settings.WASM_EXPORTS or '__main_argc_argv' in settings.WASM_EXPORTS
   if settings.HAS_MAIN and not settings.MINIMAL_RUNTIME:
@@ -234,10 +233,12 @@ def report_missing_exports(js_symbols):
     # maximum compatibility.
     return
 
-  if settings.EXPECT_MAIN and 'main' not in settings.WASM_EXPORTS and '__main_argc_argv' not in settings.WASM_EXPORTS:
-    # For compatibility with the output of wasm-ld we use the same wording here in our
-    # error message as if wasm-ld had failed.
-    exit_with_error('entry symbol not defined (pass --no-entry to suppress): main')
+  if settings.EXPECT_MAIN:
+    all_exports = settings.WASM_EXPORTS + settings.SIDE_MODULE_EXPORTS
+    if 'main' not in all_exports and '__main_argc_argv' not in all_exports:
+      # For compatibility with the output of wasm-ld we use the same wording here in our
+      # error message as if wasm-ld had failed.
+      exit_with_error('entry symbol not defined (pass --no-entry to suppress): main')
 
 
 # Test if the parentheses at body[openIdx] and body[closeIdx] are a match to
@@ -270,30 +271,6 @@ def trim_asm_const_body(body):
   return body
 
 
-def create_data_exports(data_exports):
-  lines = []
-  for k, v in data_exports.items():
-    if shared.is_internal_global(k):
-      continue
-
-    v = int(v)
-
-    if not settings.MEMORY64:
-      # We assume that global exports are addresses, which need to be interpreted as unsigned.
-      # This is not necessary (and does not work) under wasm64 when the globals are i64.
-      v = v & 0xFFFFFFFF
-
-    if settings.RELOCATABLE:
-      v += settings.GLOBAL_BASE
-    mangled = asmjs_mangle(k)
-    if should_export(mangled) and not settings.MINIMAL_RUNTIME:
-      lines.append("var %s = Module['%s'] = %s;" % (mangled, mangled, v))
-    else:
-      lines.append("var %s = %s;" % (mangled, v))
-
-  return '\n'.join(lines)
-
-
 def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadata=None):
   # Overview:
   #   * Run wasm-emscripten-finalize to extract metadata and modify the binary
@@ -374,8 +351,8 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
     if settings.INITIAL_TABLE == -1:
       settings.INITIAL_TABLE = dylink_sec.table_size + 1
 
-    if settings.ASYNCIFY == 1:
-      metadata.imports += ['__asyncify_state', '__asyncify_data']
+  if settings.MAIN_MODULE and settings.ASYNCIFY == 1:
+    metadata.imports += ['__asyncify_state', '__asyncify_data']
 
   if metadata.invoke_funcs:
     settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE += ['$getWasmTableEntry']
@@ -407,13 +384,11 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
   if base_metadata:
     function_exports = base_metadata.function_exports
     other_exports = base_metadata.other_exports
-    # We want the real values from the final metadata but we only want to
-    # include names from the base_metadata.  See phase_link() in link.py.
-    data_exports = {k: v for k, v in metadata.data_exports.items() if k in base_metadata.data_exports}
   else:
     function_exports = metadata.function_exports
     other_exports = metadata.other_exports
-    data_exports = metadata.data_exports
+
+  other_exports = [e for e in other_exports if not shared.is_internal_global(e[0].name)]
 
   if settings.ASYNCIFY == 1:
     function_exports['asyncify_start_unwind'] = webassembly.FuncType([webassembly.Type.I32], [])
@@ -422,7 +397,7 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
     function_exports['asyncify_stop_rewind'] = webassembly.FuncType([], [])
 
   parts = [pre]
-  parts += create_module(metadata, function_exports, data_exports, other_exports,
+  parts += create_module(metadata, function_exports, other_exports,
                          forwarded_json['librarySymbols'], forwarded_json['nativeAliases'])
   parts.append(post)
   settings.ALIASES = list(forwarded_json['nativeAliases'].keys())
@@ -900,9 +875,10 @@ def install_debug_wrapper(sym):
   # runtimeInitialized.
   if sym.startswith(('__asan_', 'emscripten_stack_', '_emscripten_stack_')):
     return False
-  # Likewise `__trap` can occur before the runtime is initialized since it is used in
-  # abort.
-  return sym != '__trap'
+  # `__trap` can occur before the runtime is initialized since it is used in abort.
+  # `emscripten_get_sbrk_ptr` can be called prior to runtime initialzaion by
+  # the dynamic linking code.
+  return sym not in ['__trap', 'emscripten_get_sbrk_ptr']
 
 
 def should_export(sym):
@@ -914,7 +890,7 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
   receiving = ['\n// Imports from the Wasm binary.']
 
   if settings.WASM_ESM_INTEGRATION:
-    exported_symbols = other_exports + list(function_exports.keys())
+    exported_symbols = [e[0].name for e in other_exports] + list(function_exports.keys())
     exports = []
     for sym in exported_symbols:
       mangled = asmjs_mangle(sym)
@@ -951,8 +927,8 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
   # function assignWasmExports(wasmExport) {
   #   _main = wasmExports["_main"];
   exports = {name: sig for name, sig in function_exports.items() if name != building.WASM_CALL_CTORS}
-  for name in other_exports:
-    exports[name] = None
+  for export, info in other_exports:
+    exports[export.name] = (export, info)
 
   mangled = [asmjs_mangle(s) for s in exports] + list(aliases.keys())
   if settings.ASSERTIONS:
@@ -983,8 +959,11 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
 
   do_module_exports = (settings.MODULARIZE or not settings.MINIMAL_RUNTIME) and settings.MODULARIZE != 'instance'
   receiving.append('\nfunction assignWasmExports(wasmExports) {')
-  for sym, sig in exports.items():
-    is_function = sig is not None
+  if settings.ASSERTIONS:
+    for sym in exports:
+      receiving.append(f"  assert(typeof wasmExports['{sym}'] != 'undefined', 'missing Wasm export: {sym}');")
+  for sym, info in exports.items():
+    is_function = type(info) == webassembly.FuncType
     mangled = asmjs_mangle(sym)
     assignment = mangled
     if generate_dyncall_assignment and is_function and sym.startswith('dynCall_'):
@@ -992,16 +971,24 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
       assignment += f" = dynCalls['{sig_str}']"
     if do_module_exports and should_export(mangled):
       assignment += f" = Module['{mangled}']"
-    if settings.ASSERTIONS:
-      receiving.append(f"  assert(typeof wasmExports['{sym}'] != 'undefined', 'missing Wasm export: {sym}');")
     if sym in alias_inverse_map:
       for target in alias_inverse_map[sym]:
         assignment += f" = {target}"
         if do_module_exports and target in settings.EXPORTED_RUNTIME_METHODS:
           assignment += f" = Module['{target}']"
     if is_function and install_debug_wrapper(sym):
-      nargs = len(sig.params)
+      nargs = len(info.params)
       receiving.append(f"  {assignment} = createExportWrapper('{sym}', {nargs});")
+    elif not is_function and info[0].kind == webassembly.ExternType.GLOBAL and not info[1].mutable:
+      if settings.LEGACY_VM_SUPPORT:
+        value = f"typeof wasmExports['{sym}'] == 'object' ? wasmExports['{sym}'].value : wasmExports['{sym}']"
+      else:
+        value = f"wasmExports['{sym}'].value"
+      if settings.MEMORY64:
+        value = f'Number({value})'
+      elif settings.CAN_ADDRESS_2GB:
+        value = f'({value}) >>> 0'
+      receiving.append(f"  {assignment} = {value};")
     else:
       receiving.append(f"  {assignment} = wasmExports['{sym}'];")
   receiving.append('}')
@@ -1009,10 +996,9 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
   return '\n'.join(receiving)
 
 
-def create_module(metadata, function_exports, data_exports, other_exports, library_symbols, aliases):
+def create_module(metadata, function_exports, other_exports, library_symbols, aliases):
   module = []
   module.append(create_receiving(function_exports, other_exports, library_symbols, aliases))
-  module.append(create_data_exports(data_exports))
 
   sending = create_sending(metadata, library_symbols)
   if settings.WASM_ESM_INTEGRATION:
@@ -1065,6 +1051,7 @@ def create_pointer_conversion_wrappers(metadata):
   mapping = {
     'sbrk': 'pP',
     '_emscripten_stack_alloc': 'pp',
+    'emscripten_get_sbrk_ptr': 'p',
     'emscripten_builtin_malloc': 'pp',
     'emscripten_builtin_calloc': 'ppp',
     'wasmfs_create_node_backend': 'pp',
@@ -1143,7 +1130,7 @@ def create_pointer_conversion_wrappers(metadata):
     'emscripten_main_runtime_thread_id': 'p',
     '_emscripten_set_offscreencanvas_size_on_thread': '_pp__',
     'fileno': '_p',
-    '_emscripten_run_callback_on_thread': '_pp_pp',
+    '_emscripten_run_callback_on_thread': '_pp_ppp',
     '_emscripten_find_dylib': 'ppppp',
   }
 
